# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_approach.ipynb (unless otherwise specified).

__all__ = ['flatten_dict', 'gen_extracted_features', 'gen_tfidfs', 'gen_bovw_similarity', 'fuzzy_LCS',
           'gen_lcs_similarity', 'fix_sims', 'sort_rankings', 'approach', 'compute_sims']

# Cell
import copy
import cv2
import multiprocessing
import pickle
import time

import numpy as np

from collections import defaultdict, OrderedDict
from itertools import combinations, combinations_with_replacement, permutations
from joblib import Parallel, delayed
from pathlib import Path
from .eval import *
from .features import *
from .model import *
from .prep import *
from tqdm.auto import tqdm
from sklearn.metrics.pairwise import cosine_similarity

# Cell
def flatten_dict(d_in, d_out, parent_key):
    for k, v in d_in.items():
        if isinstance(v, dict):
            flatten_dict(v, d_out, parent_key + (k,))
        else:
            d_out[parent_key + (k,)] = v

# Cell
def gen_extracted_features(vid_ds, mdl, fps, ftk):
    vid_ds_features = {}
    for app in tqdm(vid_ds.labels):
        start = time.time()
        vid_ds_features[app] = {}
        for bug in vid_ds[app]:
            vid_ds_features[app][bug] = {}
            for report in vid_ds[app][bug]:
                vid_ds_features[app][bug][report] = {
                    'features': extract_features(vid_ds[app][bug][report], mdl, fps, frames_to_keep = ftk)
                }
        end = time.time()
        vid_ds_features[app]['elapsed_time'] = end - start

    return vid_ds_features

# Cell
def gen_tfidfs(vid_ds_features, vw, codebook, df, ftk):
    vid_tfids = defaultdict(
        lambda: defaultdict(dict)
    )

    for app in vid_ds_features:
        for bug in vid_ds_features[app]:
            if bug == 'elapsed_time': continue
            for report in vid_ds_features[app][bug]:
                bovw = new_get_bovw(
                    vid_ds_features[app][bug][report]['features'],
                    codebook, vw
                )
                vid_tfids[app][bug][report] = calc_tf_idf(bovw, df)

    return vid_tfids

# Cell
def gen_bovw_similarity(vid_ds, vid_ds_features, mdl, codebook, vw, ftk):
    results = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(
                        lambda: defaultdict(float)
                    )
                )
            )
        )
    )

    vid_ds_features = copy.deepcopy(vid_ds_features)
    df = np.histogram(codebook.labels_, bins = range(vw + 1))[0]
    vid_tfids = gen_tfidfs(vid_ds_features, vw, codebook, df, ftk)
    for app, bugs in vid_ds.labels.items():
        start = time.time()
        l = [(bug, report) for bug in bugs for report in bugs[bug] if bug != 'elapsed_time']
        pairs = list(x for x in combinations_with_replacement(l, 2) if x[0] != x[1])
        for (bug_i, report_i), (bug_j, report_j) in pairs:
            results[app][bug_i][report_i][bug_j][report_j]['bovw'] = np.dot(vid_tfids[app][bug_i][report_i], vid_tfids[app][bug_j][report_j]) / (np.linalg.norm(vid_tfids[app][bug_i][report_i]) * np.linalg.norm(vid_tfids[app][bug_j][report_j]))
        end = time.time()
        results[app]['elapsed_time'] = end - start + vid_ds_features[app]['elapsed_time']

    return df, results

# Cell
# Modified from geeksforgeeks: https://www.geeksforgeeks.org/longest-common-substring-dp-29/
def fuzzy_LCS(X, Y, m, n, sim_func, codebook, df, vw, mdl_frame_threshold = 0.0):
    LCSuff = [[0 for k in range(n + 1)] for l in range(m + 1)]
    LCSuff_weighted = [[0 for k in range(n + 1)] for l in range(m + 1)]

    # To store the length of
    # longest common substring
    result = result_weighted = 0

    # Following steps to build
    # LCSuff[m+1][n+1] in bottom up fashion
    for i in range(0, m + 1):
        for j in range(0, n + 1):
            if (i == 0 or j == 0):
                LCSuff[i][j] = 0
                LCSuff_weighted[i][j] = 0
                continue

            sim = sim_func(X[i - 1], Y[j - 1], codebook, df, vw)
            if sim > mdl_frame_threshold:
                LCSuff[i][j] = LCSuff[i - 1][j - 1] + sim
                LCSuff_weighted[i][j] = LCSuff_weighted[i - 1][j - 1] + sim * (i / m) * (j / n)
                if LCSuff[i][j] > result:
                    result = LCSuff[i][j]
                    result_weighted = LCSuff_weighted[i][j]
            else:
                LCSuff[i][j] = 0
                LCSuff_weighted[i][j] = 0

    mini, maxi = min(m, n), max(m, n)
    sum_w = 0
    max_v = maxi + 1
    for i in reversed(range(1, mini + 1)):
        sum_w += (i / mini) * (max_v / maxi)
        max_v -= 1
    return result / min(m, n), result_weighted / sum_w

# Cell
def gen_lcs_similarity(vid_ds, vid_ds_features, sim_func, mdl, codebook, df, vw, ftk):
    results = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(
                        lambda: defaultdict(float)
                    )
                )
            )
        )
    )

    vid_ds_features = copy.deepcopy(vid_ds_features)
    for app, bugs in vid_ds.labels.items():
        start = time.time()
        l = [(bug, report) for bug in bugs for report in bugs[bug] if bug != 'elapsed_time']
        pairs = list(x for x in combinations_with_replacement(l, 2) if x[0] != x[1])
        for (bug_i, report_i), (bug_j, report_j) in tqdm(pairs):
            lcs_sim, weighted_lcs_sim = fuzzy_LCS(
                vid_ds_features[app][bug_i][report_i]['features'],
                vid_ds_features[app][bug_j][report_j]['features'],
                len(vid_ds_features[app][bug_i][report_i]['features']),
                len(vid_ds_features[app][bug_j][report_j]['features']),
                sim_func, codebook, df, vw
            )
            results[app][bug_i][report_i][bug_j][report_j]['lcs'] = lcs_sim
            results[app][bug_i][report_i][bug_j][report_j]['weighted_lcs'] = weighted_lcs_sim

        end = time.time()
        results[app]['elapsed_time'] = end - start + vid_ds_features[app]['elapsed_time']

    return results

# Cell
def fix_sims(vid_sims, vid_ds):
    for sim_type in vid_sims:
        for app in vid_sims[sim_type]:
            l = [(bug, report) for bug in vid_ds[app] for report in vid_ds[app][bug] if bug != 'elapsed_time']
            pairs = reversed(list(x for x in permutations(l, 2) if x[0] != x[1]))
            for (bug_i, report_i), (bug_j, report_j) in pairs:
                if (bug_i, report_i) == (bug_j, report_j): continue
                vid_sims[sim_type][app][bug_i][report_i][bug_j][report_j] = vid_sims[sim_type][app][bug_j][report_j][bug_i][report_i]

    return vid_sims

# Cell
def sort_rankings(vid_sims):
    sorted_rankings = {}
    for sim_type in vid_sims:
        sorted_rankings[sim_type] = {}
        for app in vid_sims[sim_type]:
            sorted_rankings[sim_type][app] = {'elapsed_time': vid_sims[sim_type][app][f'elapsed_time']}
            for bug in vid_sims[sim_type][app]:
                if bug == 'elapsed_time': continue
                sorted_rankings[sim_type][app][bug] = {}
                for report in vid_sims[sim_type][app][bug]:
                    sorted_rankings[sim_type][app][bug][report] = []
                    d_out = {}
                    flatten_dict(vid_sims[sim_type][app][bug][report], d_out, tuple())
                    sorted_rankings[sim_type][app][bug][report] = OrderedDict(
                        sorted(d_out.items(), key = lambda x: str(x[1]), reverse = True)
                    )

    return sorted_rankings

# Cell
def approach(
    vid_ds, vid_ds_features, bovw_vid_ds_sims, lcs_vid_ds_sims,
    mdl, sim_func, codebook, df, vw, fps = 30, ftk = 1
):
    vid_ds_sims = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(
                lambda: defaultdict(
                    lambda: defaultdict(
                        lambda: defaultdict(
                            lambda: defaultdict(float)
                        )
                    )
                )
            )
        )
    )

    vid_ds_features = copy.deepcopy(vid_ds_features)
    bovw_vid_ds_sims = copy.deepcopy(bovw_vid_ds_sims)
    lcs_vid_ds_sims = copy.deepcopy(lcs_vid_ds_sims)
    for app, bugs in vid_ds.labels.items():
        l = [(bug, report) for bug in bugs for report in bugs[bug] if bug != 'elapsed_time']
        pairs = list(x for x in combinations_with_replacement(l, 2) if x[0] != x[1])
        for (bug_i, report_i), (bug_j, report_j) in tqdm(pairs):
            lcs = lcs_vid_ds_sims[app][bug_i][report_i][bug_j][report_j]['lcs']
            weighted_lcs = lcs_vid_ds_sims[app][bug_i][report_i][bug_j][report_j]['weighted_lcs']
            vid_ds_sims['lcs'][app][bug_i][report_i][bug_j][report_j] = lcs
            vid_ds_sims['weighted_lcs'][app][bug_i][report_i][bug_j][report_j] = weighted_lcs

            bovw = bovw_vid_ds_sims[app][bug_i][report_i][bug_j][report_j]['bovw']
            vid_ds_sims['bovw'][app][bug_i][report_i][bug_j][report_j] = bovw
            vid_ds_sims['bovw_lcs'][app][bug_i][report_i][bug_j][report_j] = (bovw + lcs) / 2
            vid_ds_sims['bovw_weighted_lcs'][app][bug_i][report_i][bug_j][report_j] = (bovw + weighted_lcs) / 2

        bovw_time = bovw_vid_ds_sims[app]['elapsed_time']
        lcs_time = lcs_vid_ds_sims[app]['elapsed_time']

        vid_ds_sims['bovw'][app]['elapsed_time'] = bovw_time
        vid_ds_sims['lcs'][app]['elapsed_time'] = lcs_time
        vid_ds_sims['weighted_lcs'][app]['elapsed_time'] = lcs_time
        vid_ds_sims['bovw_lcs'][app]['elapsed_time'] = bovw_time + lcs_time
        vid_ds_sims['bovw_weighted_lcs'][app]['elapsed_time'] = bovw_time + lcs_time

    fixed_vid_ds_sims = fix_sims(vid_ds_sims, vid_ds)
    rankings = sort_rankings(fixed_vid_ds_sims)
    return rankings

# Cell
def compute_sims(q_vid, vid_ds, model, codebook, vw, fps, ftk):
    df = np.histogram(codebook.labels_, bins = range(vw + 1))[0]

    q_features = extract_features(q_vid, model, fps, frames_to_keep = ftk)
    bovw = new_get_bovw(
        q_features,
        codebook, vw
    )
    q_tfids = calc_tf_idf(bovw, df)

    vid_ds_features = gen_extracted_features(vid_ds, model, fps, ftk)
    vid_ds_tfids = gen_tfidfs(vid_ds_features, vw, codebook, df, ftk)
    results = {}
    for app in tqdm(vid_ds.labels):
        start = time.time()
        results[app] = {}
        for bug in vid_ds[app]:
            results[app][bug] = {}
            for report in vid_ds[app][bug]:
                results[app][bug][report] = np.dot(q_tfids, vid_ds_tfids[app][bug][report]) / (np.linalg.norm(q_tfids) * np.linalg.norm(vid_ds_tfids[app][bug][report]))

    d_out = {}
    flatten_dict(results, d_out, tuple())
    sorted_rankings = OrderedDict(
        sorted(d_out.items(), key = lambda x: str(x[1]), reverse = True)
    )

    return sorted_rankings


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_eval.ipynb (unless otherwise specified).

__all__ = ['calc_tf_idf', 'cosine_similarity', 'hit_rate_at_k', 'mean_reciprocal_rank', 'r_precision', 'precision_at_k',
           'average_precision', 'mean_average_precision', 'recall_at_k', 'rank_stats', 'evaluate', 'get_eval_results',
           'evaluate_ranking']

# Cell
import cv2
import ffmpeg
import pickle

import numpy as np

from fastprogress.fastprogress import progress_bar
from matplotlib import pyplot as plt
from pathlib import Path
from .prep import *
from sklearn.cluster import KMeans

# Cell
def calc_tf_idf(tfs, dfs):
    tf_idf = np.array([])
    for tf, df in zip(tfs, dfs):
        tf = tf / np.sum(tfs)
        idf = np.log(len(tfs) / (df + 1))
        tf_idf = np.append(tf_idf, tf * idf)

    return tf_idf

# Cell
def cosine_similarity(a, b):
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

# Cell
def hit_rate_at_k(rs, k):
    hits = 0
    for r in rs:
        if np.sum(r[:k]) > 0: hits += 1

    return hits / len(rs)

# Cell
def mean_reciprocal_rank(rs):
    """Score is reciprocal of the rank of the first relevant item

    First element is 'rank 1'.  Relevance is binary (nonzero is relevant).

    Example from http://en.wikipedia.org/wiki/Mean_reciprocal_rank
    >>> rs = [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.61111111111111105
    >>> rs = np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0]])
    >>> mean_reciprocal_rank(rs)
    0.5
    >>> rs = [[0, 0, 0, 1], [1, 0, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.75

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean reciprocal rank
    """
    rs = (np.asarray(r).nonzero()[0] for r in rs)
    return np.mean([1. / (r[0] + 1) if r.size else 0. for r in rs])

def r_precision(r):
    """Score is precision after all relevant documents have been retrieved

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> r_precision(r)
    0.33333333333333331
    >>> r = [0, 1, 0]
    >>> r_precision(r)
    0.5
    >>> r = [1, 0, 0]
    >>> r_precision(r)
    1.0

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        R Precision
    """
    r = np.asarray(r) != 0
    z = r.nonzero()[0]
    if not z.size:
        return 0.
    return np.mean(r[:z[-1] + 1])


def precision_at_k(r, k):
    """Score is precision @ k

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> precision_at_k(r, 1)
    0.0
    >>> precision_at_k(r, 2)
    0.0
    >>> precision_at_k(r, 3)
    0.33333333333333331
    >>> precision_at_k(r, 4)
    Traceback (most recent call last):
        File "<stdin>", line 1, in ?
    ValueError: Relevance score length < k


    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Precision @ k

    Raises:
        ValueError: len(r) must be >= k
    """
    assert k >= 1
    r = np.asarray(r)[:k] != 0
    if r.size != k:
        raise ValueError('Relevance score length < k')
    return np.mean(r)


def average_precision(r):
    """Score is average precision (area under PR curve)

    Relevance is binary (nonzero is relevant).

    >>> r = [1, 1, 0, 1, 0, 1, 0, 0, 0, 1]
    >>> delta_r = 1. / sum(r)
    >>> sum([sum(r[:x + 1]) / (x + 1.) * delta_r for x, y in enumerate(r) if y])
    0.7833333333333333
    >>> average_precision(r)
    0.78333333333333333

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Average precision
    """
    r = np.asarray(r) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.
    return np.mean(out)


def mean_average_precision(rs):
    """Score is mean average precision

    Relevance is binary (nonzero is relevant).

    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1]]
    >>> mean_average_precision(rs)
    0.78333333333333333
    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1], [0]]
    >>> mean_average_precision(rs)
    0.39166666666666666

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean average precision
    """
    return np.mean([average_precision(r) for r in rs])

def recall_at_k(r, k, l):
    """Score is recall @ k

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> recall_at_k(r, 1, 2)
    0.0
    >>> recall_at_k(r, 2, 2)
    0.0
    >>> recall_at_k(r, 3, 2)
    0.5
    >>> recall_at_k(r, 4, 2)
    Traceback (most recent call last):
        File "<stdin>", line 1, in ?
    ValueError: Relevance score length < k


    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
        k: the length or size of the relevant items

    Returns:
        Recall @ k

    Raises:
        ValueError: len(r) must be >= k
    """
    assert k >= 1
    assert l >= 1
    r2 = np.asarray(r)[:k] != 0
    if r2.size != k:
        raise ValueError('Relevance score length < k')
    return np.sum(r2)/l

# Cell
def rank_stats(rs):
    ranks = []
    for r in rs:
        ranks.append(r.nonzero()[0][0] + 1)
    ranks = np.asarray(ranks)
    recipical_ranks = 1 / ranks
    return np.std(ranks), np.mean(ranks), np.median(ranks), np.mean(recipical_ranks)

# Cell
def evaluate(rankings, top_k = [1, 5, 10]):
    output = {}
    for app in rankings:
        output[app] = {}
        app_rs = []
        for bug in rankings[app]:
            if bug == 'elapsed_time': continue
            output[app][bug] = {}
            bug_rs = []
            for report in rankings[app][bug]:
                output[app][bug][report] = {'ranks': []}
                r = []
                for labels, score in rankings[app][bug][report].items():
                    output[app][bug][report]['ranks'].append((labels, score))
                    if labels[0] == bug: r.append(1)
                    else: r.append(0)
                r = np.asarray(r)
                output[app][bug][report]['rank'] = r.nonzero()[0][0] + 1
                output[app][bug][report]['average_precision'] = average_precision(r)
                bug_rs.append(r)

            bug_rs_std, bug_rs_mean, bug_rs_med, bug_mRR = rank_stats(bug_rs)
            bug_mAP = mean_average_precision(bug_rs)

            output[app][bug]['Bug std rank'] = bug_rs_std
            output[app][bug]['Bug mean rank'] = bug_rs_mean
            output[app][bug]['Bug median rank'] = bug_rs_med
            output[app][bug]['Bug mRR'] = bug_mRR
            output[app][bug]['Bug mAP'] = bug_mAP
            for k in top_k:
                bug_hit_rate = hit_rate_at_k(bug_rs, k)
                output[app][f'Bug Hit@{k}'] = bug_hit_rate
            app_rs.extend(bug_rs)

        app_rs_std, app_rs_mean, app_rs_med, app_mRR = rank_stats(app_rs)
        app_mAP = mean_average_precision(app_rs)

        output[app]['App std rank'] = app_rs_std
        output[app]['App mean rank'] = app_rs_mean
        output[app]['App median rank'] = app_rs_med
        output[app]['App mRR'] = app_mRR
        output[app]['App mAP'] = app_mAP
        print(f'{app} Elapsed Time in Seconds', rankings[app]['elapsed_time'])
        print(f'{app} σ Rank', app_rs_std)
        print(f'{app} μ Rank', app_rs_mean)
        print(f'{app} Median Rank', app_rs_med)
        print(f'{app} mRR:', app_mRR)
        print(f'{app} mAP:', app_mAP)
        for k in top_k:
            app_hit_rate = hit_rate_at_k(app_rs, k)
            output[app][f'App Hit@{k}'] = app_hit_rate
            print(f'{app} Hit@{k}:', app_hit_rate)

    return output

# Cell
def get_eval_results(evals, app, item):
    for bug in evals[app]:
        if bug == 'elapsed_time': continue
        for vid in evals[app][bug]:
            try:
                print(evals[app][bug][vid][item])
            except: continue

# Cell
def evaluate_ranking(ranking, ground_truth):
    relevance = []
    for doc in ranking:
        if doc in ground_truth:
            relevance.append(1)
        else:
            relevance.append(0)

    r = np.asarray(relevance)
    first_rank = int(r.nonzero()[0][0] + 1)
    avg_precision = average_precision(r)
    recip_rank = 1 / first_rank

    ranks = []
    precisions = []
    recalls = []

    limit = 10

    for k in range(1, limit + 1):
        ranks.append(1 if first_rank <= k else 0)
        precisions.append(precision_at_k(r, k))
        recalls.append(recall_at_k(r, k, len(ground_truth)))

    results = {
        'first_rank': first_rank,
        'recip_rank': recip_rank,
        'avg_precision': avg_precision
    }

    for i in range(limit):
        k = i + 1
        results["rr@" + str(k)] = ranks[i]
        results["p@" + str(k)] = precisions[i]
        results["r@" + str(k)] = recalls[i]

    return results


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_cli.ipynb (unless otherwise specified).

__all__ = ['logger', 'URLs', 'download', 'VWORDS', 'N_IMGS', 'N_FRAMES_TO_KEEP', 'FPS', 'BEST_DL_MODELS',
           'BEST_IR_MODELS', 'BEST_MODEL_CONFIGS', 'reproduce', 'tango']

# Cell
import logging
import io
import pickle
import pprint
import random
import requests
import subprocess
import time
import zipfile

import numpy as np
import pandas as pd

from fastcore.script import call_parse, Param
from pathlib import Path
from .prep import *
from .features import *
from .eval import *
from .model import *
from .approach import *
from .combo import *
from tqdm.auto import tqdm

logger = logging.getLogger()
logger.setLevel(logging.INFO)

# Cell
URLs = {
    "tango_reproduction_package": "https://zenodo.org/record/4453765/files/tango_reproduction_package.zip",
}

# Cell
# @call_parse
def _download(
    out_path
):
    """Function for downloading all data and results related to this tool's paper"""
    out_path = Path(out_path)
    out_path.mkdir(parents=True, exist_ok=True)

    logging.info(f"Downloading and extracting datasets and models to {str(out_path)}.")
    r = requests.get(URLs["tango_reproduction_package"])
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(out_path)

# Cell
@call_parse
def download(
    out_path: Param("The output path to save and unzip all files.", str)
):
    _download(out_path)

# Cell
# all hyperparameters used
VWORDS = [1_000, 5_000, 10_000]
N_IMGS = 15_000
N_FRAMES_TO_KEEP = [1, 5]
FPS = 30

# Fix naming issue and number of models reported...
BEST_DL_MODELS= [
    "SIFT-10000vw-1ftk-bovw_weighted_lcs",
    "SimCLR-1000vw-5ftk-bovw", "SimCLR-5000vw-5ftk-bovw_lcs",
    "SimCLR-5000vw-5ftk-bovw_weighted_lcs", "SimCLR-1000vw-5ftk-bovw_weighted_lcs"
]
BEST_IR_MODELS = [
    "ocr+ir-1ftk-all_text", "ocr+ir-5ftk-all_text",
    "ocr+ir-5ftk-unique_frames", "ocr+ir-5ftk-unique_words"
]

BEST_MODEL_CONFIGS = {
    "SimCLR": "SimCLR-1000vw-5ftk-bovw",
    "SIFT": "SIFT-10000vw-1ftk-bovw_weighted_lcs",
    "OCR+IR": "ocr+ir-5ftk-all_text"
}

# Cell
def _generate_vis_results(vid_ds, out_path, art_path, vis_model):
    if vis_model == "SimCLR":
        simclr = SimCLRModel.load_from_checkpoint(
            checkpoint_path = str(
                art_path/"models"/"SimCLR"/"checkpointepoch=98.ckpt"
            )
        ).eval()
        model = SimCLRExtractor(simclr)
        sim_func = simclr_frame_sim
    else:
        model = SIFTExtractor(cv2.xfeatures2d.SIFT_create(nfeatures = 10))
        sim_func = sift_frame_sim

    logging.info(f"Computing rankings and calculating metrics for {vis_model} visual model.")
    for vw in tqdm(VWORDS):
        for ftk in tqdm(N_FRAMES_TO_KEEP):
            evaluation_metrics = {}
            cb_path = art_path/"models"/vis_model/f"cookbook_{vis_model}_{vw}vw.model"
            codebook = pickle.load(open(cb_path, "rb"))
            start = time.time()
            vid_ds_features = gen_extracted_features(vid_ds, model, FPS, ftk)
            end = time.time()
            feature_gen_time = end - start

            df, bovw_vid_ds_sims = gen_bovw_similarity(
                vid_ds, vid_ds_features, model, codebook, vw, ftk
            )
            lcs_vid_ds_sims = gen_lcs_similarity(
                vid_ds, vid_ds_features, sim_func, model, codebook, df, vw, ftk
            )
            rankings = approach(
                vid_ds, vid_ds_features, bovw_vid_ds_sims, lcs_vid_ds_sims, model, sim_func,
                codebook, df, vw, fps = FPS, ftk = ftk,
            )

            for k, v in rankings.items():
                evaluation_metrics[k] = evaluate(rankings[k])

            id_name = f"user_{N_IMGS}n_{vw}vw_{FPS}fps_{ftk}ftk"
            results_path = out_path/"results"/vis_model
            results_path.mkdir(parents=True, exist_ok=True)
            logging.info(f"Saving rankings and metrics to {str(results_path)}.")
            with open(results_path/f"rankings_{id_name}.pkl", "wb") as f:
                pickle.dump(rankings, f, protocol=pickle.HIGHEST_PROTOCOL)

            with open(results_path/f"evaluation_metrics_{id_name}.pkl", 'wb') as f:
                pickle.dump(evaluation_metrics, f, protocol=pickle.HIGHEST_PROTOCOL)

# Cell
def _generate_txt_results(vid_ds, out_path, art_path, vis_model):
    logging.info("Computing rankings and calculating metrics for textual model.")
    csv_file_path = art_path/"user_assignment.csv"
    settings_path = out_path/"evaluation_settings"
    settings_path.mkdir(parents=True, exist_ok=True)
    video_data = read_video_data(csv_file_path)
    generate_setting2(video_data, settings_path)
    convert_results_format(out_path/"results", settings_path, out_path, [vis_model])

    # Check if files already exist and skip if they do because it takes a long time
    txt_out_path = out_path/"extracted_text"
    for ftk in N_FRAMES_TO_KEEP:
        if not (txt_out_path/f"text_{ftk}").exists():
            get_all_texts(vid_ds, txt_out_path, fps = ftk)

    txt_path = art_path/"models"/"OCR+IR"
    subprocess.check_output(
        ["sh", "build_run.sh", str(txt_out_path), str(settings_path)],
        cwd=str(txt_path),
    )

# Cell
def _get_single_model_performance(pkl_path, technique):
    evals = pickle.load(open(fname, 'rb'))
    mRRs = []
    mAPs = []
    mean_Rs = []
    for app in evals[technique]:
        mRRs.append(evals[technique][app]["App mRR"])
        mAPs.append(evals[technique][app]["App mAP"])
        mean_Rs.append(evals[technique][app]["App mean rank"])

    return np.mean(mRRs), np.mean(mAPs), np.mean(mean_Rs)

# Cell
def _print_performance(model, mRR, mAP, mean_R):
    print(
        f"""\
        Model: {model}
        Overall mRR: {mRR}
        Overall mAP: {mAP}
        Overall Mean Rank: {mean_R}
        """
    )
def _output_performance(out_path, dl_model_config, ir_model_config):
    all_results = pd.read_csv(
        out_path/"combined"/"tango_comb_results"/"all_results.csv", sep=";"
    )
    # Print the comb model results
    comb_results = all_results[
        all_results["model_config"] == f"({dl_model_config},{ir_model_config})"
    ][all_results["weight"] == "0.2-0"]
    _print_performance(
        f"{dl_model_config},{ir_model_config},weight=0.2-0", np.mean(comb_results['recip_rank'].values),
        np.mean(comb_results['avg_precision'].values), np.mean(comb_results['first_rank'].values)
    )

    # Print the vis model results
    vis_singl_results = all_results[
        all_results["model_config"] == f"({dl_model_config},{ir_model_config})"
    ][all_results["weight"] == "0.0"]
    _print_performance(
        f"{dl_model_config}", np.mean(vis_singl_results['recip_rank'].values),
        np.mean(vis_singl_results['avg_precision'].values), np.mean(vis_singl_results['first_rank'].values)
    )
    # Print the txt model results
    ir_singl_results = all_results[
        all_results["model_config"] == f"({dl_model_config},{ir_model_config})"
    ][all_results["weight"] == "1.0"]
    _print_performance(
        f"{ir_model_config}", np.mean(ir_singl_results['recip_rank'].values),
        np.mean(ir_singl_results['avg_precision'].values), np.mean(ir_singl_results['first_rank'].values)
    )

# Cell
@call_parse
def reproduce(
    down_path: Param("The directory where all the files will be downloaded and extracted to.", str),
    out_path: Param("The output path to place all results in.", str),
    vis_model: Param("The type of visual model. Can be either SimCLR or SIFT, taking ~6 hours or >2 weeks, respectively, for all apps on our machine with 755G of RAM and 72 CPUs.", str)
):
    """Function for reproducing all results related to this tool's paper"""
    random.seed(42)
    _download(down_path)
    down_path = Path(down_path)
    out_path = Path(out_path)
    art_path = down_path/"tango_reproduction_package"/"artifacts"

    logging.info("Loading videos.")
    vid_ds = VideoDataset.from_path(
        art_path/"videos", fr = FPS
    ).label_from_paths()

    _generate_vis_results(vid_ds, out_path, art_path, vis_model)
    _generate_txt_results(vid_ds, out_path, art_path, vis_model)

    combo_out_path = out_path/"combined"
    dl_ranking_path = out_path/"user_rankings_weighted_all"/"all_rankings.csv"
    txt_path = art_path/"models"/"OCR+IR"
    ir_rankings_path = txt_path/"tango_txt_rankings"/"all_rankings.json"
    settings_path = out_path/"evaluation_settings"

    tango_combined(combo_out_path, dl_ranking_path, ir_rankings_path, settings_path, BEST_DL_MODELS, BEST_IR_MODELS)
    _output_performance(out_path, BEST_MODEL_CONFIGS[vis_model], BEST_MODEL_CONFIGS["OCR+IR"])

# Cell
@call_parse
def tango(
    q_path: Param("Path to the query video", str),
    cor_path: Param("Path to the corpus", str),
    simclr_path: Param("Path to the SimCLR model directory", str)
):
    """
    Function for calculating similarity scores of a corpus of video-based bug
    reports to a query video-based bug report. Currently only uses the top
    performing SimCLR model from our paper
    "It Takes Two to TANGO: Combining Visual andTextual Information for Detecting DuplicateVideo-Based Bug Reports".
    """
    q_path = Path(q_path)
    cor_path = Path(cor_path)
    simclr_path = Path(simclr_path)
    best_vw = 1_000
    best_ftk = 5

    q_vid = Video(q_path, FPS)
    codebook = pickle.load(open(simclr_path/f"cookbook_SimCLR_{best_vw}vw.model", 'rb'))
    simclr = SimCLRModel.load_from_checkpoint(
        checkpoint_path=str(simclr_path/"checkpointepoch=98.ckpt")
    ).eval()
    model = SimCLRExtractor(simclr)

    vid_ds = VideoDataset.from_path(cor_path, fr=FPS).label_from_paths()
    sorted_rankings = compute_sims(q_vid, vid_ds, model, codebook, best_vw, FPS, best_ftk)
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(sorted_rankings)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_model.ipynb (unless otherwise specified).

__all__ = ['sift_frame_sim', 'simclr_frame_sim', 'SimCLRDataset', 'imagenet_normalize_transform',
           'get_train_transforms', 'get_val_transforms', 'NTXEntCriterion', 'SimCLRModel']

# Cell
import cv2
import torch
import torchvision

import numpy as np
import torch.nn as nn

from pytorch_lightning import Trainer
from pytorch_lightning.core.lightning import LightningModule
from .features import *
from torch.autograd import Variable
from torch.utils.data import DataLoader, Dataset
from torchvision import datasets, models, transforms
from sklearn.metrics.pairwise import cosine_similarity

# Cell
def sift_frame_sim(features_i, features_j, codebook, df, vw):
    vws_i = np.expand_dims(codebook.predict(features_i), axis=0)
    bowv_i = np.expand_dims(np.histogram(vws_i, bins = range(vw + 1))[0], axis=0)
    tf_idf_i = calc_tf_idf(bowv_i, df)

    vws_j = np.expand_dims(codebook.predict(features_j), axis=0)
    bowv_j = np.expand_dims(np.histogram(vws_j, bins = range(vw + 1))[0], axis=0)
    tf_idf_j = calc_tf_idf(bowv_j, df)

    sim = np.dot(tf_idf_i, tf_idf_j) / (np.linalg.norm(tf_idf_i) * np.linalg.norm(tf_idf_j))
    return sim

# Cell
def simclr_frame_sim(features_i, features_j, codebook, df, vw):
    sim = cosine_similarity(features_i, features_j)[0][0]
    return sim

# Cell

"""
All code from: https://github.com/dthiagarajan/simclr_pytorch
Only included directory for easy loading of checkpoints
"""
class SimCLRDataset(Dataset):
    def __init__(self, dataset):
        """Initialize a wrapper of a generic image classification dataset for SimCLR training.

        Args:
            dataset (torch.utils.data.Dataset): an image PyTorch dataset - when iterating over it
                it should return something of the form (image) or (image, label).
        """
        self.dataset = dataset

    def __getitem__(self, index):
        dataset_item = self.dataset[index]
        if type(dataset_item) is tuple:
            image = dataset_item[0]
        else:
            image = dataset_item
        return image, image

    def __len__(self):
        return len(self.dataset)

    @staticmethod
    def mixup(x, alpha=0.4):
        batch_size = x.size()[0] // 2
        if alpha > 0:
            lam = np.random.beta(alpha, alpha, batch_size)
            lam = np.concatenate(
                [lam[:, None], 1 - lam[:, None]], 1
            ).max(1)[:, None, None, None]
            lam = torch.from_numpy(lam).float()
            if torch.cuda.is_available():
                lam = lam.cuda()
        else:
            lam = 1.
        # This is SimCLR specific - we want to use the same mixing for the augmented pairs
        lam = torch.cat([lam, lam])
        index = torch.randperm(batch_size)
        # This is SimCLR specific - we want to use the same permutation on the augmented pairs
        index = torch.cat([index, batch_size + index])
        if torch.cuda.is_available():
            index = index.cuda()
        mixed_x = lam * x + (1 - lam) * x[index, :]

        return mixed_x, lam


def imagenet_normalize_transform():
    return transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])


def get_train_transforms(size=224, color_jitter_prob=0.8, grayscale_prob=0.2):
    color_jitter = transforms.ColorJitter(0.8, 0.8, 0.8, 0.2)
    return transforms.Compose([
        transforms.RandomResizedCrop(size=(size, size)),
        transforms.RandomHorizontalFlip(),
        transforms.RandomApply([color_jitter], p=color_jitter_prob),
        transforms.RandomGrayscale(p=grayscale_prob),
        transforms.ToTensor(),
        imagenet_normalize_transform()
    ])


def get_val_transforms(size=224):
    return transforms.Compose([
        transforms.Resize(size=(size, size)),
        transforms.ToTensor(),
        imagenet_normalize_transform()
    ])


import torch.nn as nn


class NTXEntCriterion(nn.Module):
    """Normalized, temperature-scaled cross-entropy criterion, as suggested in the SimCLR paper.

    Parameters:
        temperature (float, optional): temperature to scale the confidences. Defaults to 0.5.
    """
    criterion = nn.CrossEntropyLoss(reduction="sum")
    similarity = nn.CosineSimilarity(dim=2)

    def __init__(self, temperature=0.5):
        super(NTXEntCriterion, self).__init__()
        self.temperature = temperature
        self.batch_size = None
        self.mask = None

    def mask_correlated_samples(self, batch_size):
        """Masks examples in a batch and it's augmented pair for computing the valid summands for
            the criterion.

        Args:
            batch_size (int): batch size of the individual batch (not including it's augmented pair)

        Returns:
            torch.Tensor: a mask (tensor of 0s and 1s), where 1s indicates a pair of examples in a
                batch that will contribute to the overall batch loss
        """
        mask = torch.ones((batch_size * 2, batch_size * 2), dtype=bool)
        mask = mask.fill_diagonal_(0)
        for i in range(batch_size):
            mask[i, batch_size + i] = 0
            mask[batch_size + i, i] = 0
        return mask

    def compute_similarities(self, z_i, z_j, temperature):
        """Computes the similarities between two projections `z_i` and `z_j`, scaling based on
            `temperature`.

        Args:
            z_i (torch.Tensor): projection of a batch
            z_j (torch.Tensor): projection of the augmented pair for the batch
            temperature (float): temperature to scale the similarity by

        Returns:
            torch.Tensor: tensor of similarities for the positive and negative pairs
        """
        batch_size = len(z_i)
        mask = self.mask_correlated_samples(batch_size)

        p1 = torch.cat((z_i, z_j), dim=0)
        sim = self.similarity(p1.unsqueeze(1), p1.unsqueeze(0)) / temperature

        sim_i_j = torch.diag(sim, batch_size)
        sim_j_i = torch.diag(sim, -batch_size)

        positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(
            batch_size * 2, 1
        )
        negative_samples = sim[mask].reshape(batch_size * 2, -1)

        logits = torch.cat((positive_samples, negative_samples), dim=1)
        return logits

    def forward(self, z):
        """Computes the loss for a batch and its augmented pair.

        Args:
            z (torch.Tensor): tensor of a batch and it's augmented pair, concatenated

        Returns:
            torch.Tensor: loss for the given batch
        """
        double_batch_size = len(z)
        batch_size = double_batch_size // 2
        z_i, z_j = z[:double_batch_size // 2], z[double_batch_size // 2:]
        if self.batch_size is None or batch_size != self.batch_size:
            self.batch_size = batch_size
            self.mask = None

        if self.mask is None:
            self.mask = self.mask_correlated_samples(self.batch_size)

        logits = self.compute_similarities(z_i, z_j, self.temperature)
        labels = torch.zeros(self.batch_size * 2).long()
        logits, labels = logits.to(z.device), labels.to(z.device)
        loss = self.criterion(logits, labels)
        loss /= 2 * self.batch_size
        return loss

class SimCLRModel(LightningModule):
    """SimCLR training network for a generic torchvision model (restricted to `allowed_models`). """

    allowed_models = ['resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
    allowed_datasets = ['CIFAR10', 'CIFAR100', 'STL10', 'SVHN']

    def __init__(
        self, model_name='resnet18', pretrained=True, projection_dim=64, temperature=0.5,
        batch_size=128, image_size=224, save_hparams=True
    ):
        super().__init__()
        layers = list(getattr(torchvision.models, model_name)(pretrained=pretrained).children())
        self.model = nn.Sequential(*layers[:-1])
        self.projection_head = nn.Linear(layers[-1].in_features, projection_dim)
        self.loss = NTXEntCriterion(temperature=temperature)
        self.batch_size = batch_size
        self.image_size = image_size
        self.prepare_data()

    def forward(self, x):
        out = self.model(x)
        out = out.view(x.size(0), -1)
        out = self.projection_head(out)
        return out

    def training_step(self, batch, batch_idx):
        projections = self(batch)
        loss = self.loss(projections)
        tensorboard_logs = {'train_loss': loss}
        self.logger.scalar('loss', loss)
        return {'loss': loss, 'log': tensorboard_logs}

    def training_epoch_end(self, outputs):
        loss_mean = torch.stack([x['loss'] for x in outputs]).mean()
        return {'train_loss': loss_mean}

    def configure_optimizers(self):
        return torch.optim.Adam([
            {'params': self.model.parameters(), 'lr': 0.00001},
            {'params': self.projection_head.parameters(), 'lr': 0.001}
        ])

    def prepare_data(self):
        pass

    def collate_fn(self, batch):
        return torch.cat([torch.stack([b[0] for b in batch]), torch.stack([b[1] for b in batch])])

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset, batch_size=self.batch_size, num_workers=64, shuffle=True,
            collate_fn=self.collate_fn
        )


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/08_combo.ipynb (unless otherwise specified).

__all__ = ['execute_retrieval_run', 'run_settings', 'write_results', 'write_rankings', 'convert_results_format',
           'get_info_to_ranking_results', 'tango_combined']

# Cell
import json
import ntpath
import os
import pickle
import sys
import time

import numpy as np
import pandas as pd

from collections import OrderedDict
from pathlib import Path
from .eval import *
from .utils import *

# Cell
def execute_retrieval_run(run, similarities):
    ranking = {}
    query = run["query"]
    corpus = run["dup_corpus"] + run["non_dup_corpus"]

    query_tokens = query.split("-")
    query_sims = similarities[query_tokens[0]][query_tokens[1]][query_tokens[2]]

    for doc in corpus:
        doc_tokens = doc.split("-")
        ranking[doc] = query_sims[(doc_tokens[1], doc_tokens[2])]

    ranking = OrderedDict(sorted(ranking.items(), key=lambda t: t[1], reverse=True))
    return ranking


def run_settings(settings, similarities, config, systems_allowed=[]):
    all_results = {}
    all_rankings = {}
    for setting in settings:
        all_results[setting] = []
        all_rankings[setting] = []

        runs = settings[setting]

        print("Running setting", setting)
        for run in runs:
            query = run["query"]
            query_tokens = query.split("-")

            if len(systems_allowed) != 0 and query_tokens[0] not in systems_allowed:
                continue

            ranking = execute_retrieval_run(run, similarities)
            ranking_results = evaluate_ranking(ranking, run["gnd_trh"])

            ranking_results["setting"] = setting
            ranking_results["app"] = query_tokens[0]
            ranking_results["run_id"] = run["run_id"]
            ranking_results.update(config)

            ranking_info = {"run_id": run["run_id"], "query": query, "ranking": ranking}
            ranking_info["setting"] = setting
            ranking_info.update(config)

            all_results[setting].append(ranking_results)
            all_rankings[setting].append(ranking_info)

    return all_results, all_rankings


def write_results(output_path, results):
    Path(output_path).mkdir(parents=True, exist_ok=True)
    all_results = []
    for setting in results:
        pd.read_json(json.dumps(results[setting])).to_csv(os.path.join(output_path, setting + '.csv'),
                                                          index=False, sep=";")
        all_results.extend(results[setting])

    pd.read_json(json.dumps(all_results)).to_csv(os.path.join(output_path, 'all_results.csv'),
                                                 index=False, sep=";")

def write_rankings(output_path, rankings):
    Path(output_path).mkdir(parents=True, exist_ok=True)
    all_rankings = []
    for setting in rankings:
        write_json_line_by_line(rankings[setting], os.path.join(output_path, setting + '.csv'))
        all_rankings.extend(rankings[setting])
    write_json_line_by_line(all_rankings, os.path.join(output_path, 'all_rankings.csv'))

# Cell
def convert_results_format(sim_path, settings_path, out_path, models):
    similarities_path = sim_path
    output_results = out_path/"user_results_weighted_all"
    output_rankings = out_path/"user_rankings_weighted_all"
    techniques = ["weighted_lcs", "bovw", "lcs", "bovw_lcs", "bovw_weighted_lcs"]
    systems_allowed = []

    settings_path = settings_path
    settings = load_settings(settings_path)

    all_results = {}
    all_rankings = {}

    for setting in settings:
        all_results[setting] = []
        all_rankings[setting] = []

    for model in models:
        sim_files = find_file("rankings_user_*.pkl", os.path.join(similarities_path, model))
        for sim_file in sim_files:
            file_name = ntpath.basename(sim_file).split(".")[0]
            file_tokens = file_name.split("_")

            vwords = file_tokens[3]
            frames_per_sec = file_tokens[4]
            ftk = file_tokens[5]

            model_similarities = pickle.load(open(sim_file, 'rb'))

            for technique in techniques:
                print(model_similarities.keys())
                similarities = model_similarities[technique]
                configuration = {
                    "model": model,
                    "vwords": vwords,
                    "fps": frames_per_sec,
                    "ftk": ftk,
                    "technique": technique
                }

                print("Running config: ", configuration)

                results, rankings = run_settings(settings, similarities, configuration, systems_allowed)

                for setting in settings:
                    all_results[setting].extend(results[setting])
                    all_rankings[setting].extend(rankings[setting])

    print("Writing results and rankings")

    write_results(output_results, all_results)
    write_rankings(output_rankings, all_rankings)

    print("done")


# Cell
def get_info_to_ranking_results(ranking, ranking_results, run, dl_model, ir_model, weight_str, setting):
    new_model = dl_model[0] + "-" + ir_model[0]
    new_vwords = dl_model[1]
    new_fps = dl_model[2] + "-" + ir_model[1] + "ftk"
    new_technique = dl_model[3] + "-" + ir_model[2]
    new_config = "({},{})".format("-".join(dl_model), "-".join(ir_model))
    new_config_weight = "({},{},{})".format(weight_str, "-".join(dl_model), "-".join(ir_model))
    config = {
        "model": new_model,
        "vwords": new_vwords,
        "fps": new_fps,
        "technique": new_technique,
        "weight": weight_str,
        "model_config": new_config,
        "model_config_weight": new_config_weight
    }

    query = run["query"]
    query_tokens = query.split("-")

    ranking_results["setting"] = setting
    ranking_results["app"] = query_tokens[0]
    ranking_results["run_id"] = run["run_id"]
    ranking_results.update(config)

    ranking_info = {"run_id": run["run_id"], "query": query, "ranking": ranking}
    ranking_info.update(config)

    return ranking_info, ranking_results

# Cell
def tango_combined(
    out_path, dl_rankings_path, ir_rankings_path,
    settings_path, dl_models, ir_models
):
    # all_data
    results_out_path = out_path/"tango_comb_results"
    rankings_out_path = out_path/"tango_comb_rankings"

    # calibration
    # settings_path = 'evaluation_settings_split/calibration'
    # results_out_path = "comb_results_calib"
    # rankings_out_path = "comb_rankings_calib"

    # test
    # settings_path = 'evaluation_settings_split/test'
    # results_out_path = "comb_results_test"
    # rankings_out_path = "comb_rankings_test"

    Path(results_out_path).mkdir(parents=True, exist_ok=True)
    Path(rankings_out_path).mkdir(parents=True, exist_ok=True)

    # read data
    settings = load_settings(settings_path)

    dl_rankings = read_json_line_by_line(dl_rankings_path)
    dl_rankings_by_config = group_dict(dl_rankings, lambda rec: (rec['model'], rec['vwords'], rec['ftk'],
                                                                    rec['technique'],))

    ir_rankings = read_json(ir_rankings_path)
    ir_rankings_by_config = group_dict(ir_rankings, lambda rec: (rec['model'], rec['fps'] + "ftk",
                                                                    rec['technique'],))

#     best_dl_models = [
#         "M00-10000vw-1ftk-bovw_weighted_lcs", "M00-10000vw-1ftk-weighted_lcs",
#         "M00-10000vw-5ftk-bovw_weighted_lcs", "M00-1000vw-1ftk-weighted_lcs", "M00-1000vw-5ftk-bovw",
#         "M00-1000vw-5ftk-bovw_weighted_lcs", "M00-5000vw-5ftk-weighted_lcs", "M01-1000vw-5ftk-bovw",
#         "M01-5000vw-5ftk-bovw_lcs", "M01-5000vw-5ftk-bovw_weighted_lcs",
#         "M01-1000vw-5ftk-bovw_weighted_lcs"]
#     best_ir_models = ["ocr+ir--1ftk-all_text", "ocr+ir--5ftk-all_text", "ocr+ir--5ftk-unique_frames",
#                       "ocr+ir--5ftk-unique_words"]

    ir_model_apps_for_comb = {
        "1ftk-all_text": ['APOD', 'DROID', 'GNU', 'GROW'],
        "5ftk-all_text": ['APOD', 'DROID', 'GNU', 'GROW'],
        "5ftk-unique_frames": ['APOD', 'DROID', 'GROW'],
        "5ftk-unique_words": ['APOD', 'GROW'],
    }

    settings_to_run = ["setting2"]

    dl_models = list(filter(lambda rec: "-".join([rec[0], rec[1], rec[2], rec[3]]) in dl_models,
                            dl_rankings_by_config.keys()))
    ir_models = list(filter(lambda rec: "-".join([rec[0], rec[1], rec[2]]) in ir_models,
                            ir_rankings_by_config.keys()))
    # run combinations

    start_time = time.time()

    all_new_rankings = []
    all_new_results = []
    for dl_model in dl_models:
        dl_mod_rankings = group_dict(dl_rankings_by_config[dl_model], lambda rec: rec["setting"])
        for ir_model in ir_models:
            ir_mod_rankings = group_dict(ir_rankings_by_config[ir_model], lambda rec: rec["setting"])

            print(dl_model, ir_model)

            app_for_comb = ir_model_apps_for_comb["-".join([ir_model[1], ir_model[2]])]

            for setting in settings_to_run:
                dl_runs = group_dict(dl_mod_rankings[setting], lambda rec: rec["run_id"])
                ir_runs = group_dict(ir_mod_rankings[setting], lambda rec: rec["runId"])

                setting_runs = settings[setting]

                for run in setting_runs:

                    run_id = run["run_id"]

                    ir_run_ranking = ir_runs[str(run_id)][0]["ranking"]
                    ir_run_ranking = dict(
                        zip((rec["docName"] for rec in ir_run_ranking), (rec for rec in ir_run_ranking)))
                    dl_run_ranking = dl_runs[run_id][0]["ranking"]

                    # rankings based on all weights
                    for weight in np.arange(0, 1.1, 0.1):
                        new_ranking = {}
                        for doc in dl_run_ranking:
                            ir_score = 0 if doc not in ir_run_ranking else ir_run_ranking[doc]["score"]
                            dl_score = dl_run_ranking[doc]
                            new_score = weight * ir_score + (1 - weight) * dl_score
                            new_ranking[doc] = new_score

                        ranking = OrderedDict(sorted(new_ranking.items(), key=lambda t: t[1], reverse=True))
                        ranking_results = evaluate_ranking(ranking, run["gnd_trh"])

                        ranking_info, ranking_results = get_info_to_ranking_results(ranking, ranking_results,
                                                                                    run, dl_model, ir_model,
                                                                                    str(weight), setting)

                        all_new_results.append(ranking_results)
                        all_new_rankings.append(ranking_info)

                # -------------------------------------------------------------

                # rankings of approach based vocabulary agreement (e.g., 0.2-0: 0.2 weight for all apps except TIME,
                # TOK, and 0 weight for TIME and TOK)
                for run in setting_runs:

                    run_id = run["run_id"]

                    ir_run_ranking = ir_runs[str(run_id)][0]["ranking"]
                    ir_run_ranking = dict(
                        zip((rec["docName"] for rec in ir_run_ranking), (rec for rec in ir_run_ranking)))
                    dl_run_ranking = dl_runs[run_id][0]["ranking"]

                    query = run["query"]
                    query_tokens = query.split("-")

                    app = query_tokens[0]

                    for base_weight in np.arange(0.1, 1.1, 0.1):

                        best_weights_name = f'{base_weight:0.1f}' + "-0"

                        weight = 0
                        if app in app_for_comb:
                            weight = base_weight

                        new_ranking = {}
                        for doc in dl_run_ranking:
                            ir_score = 0 if doc not in ir_run_ranking else ir_run_ranking[doc]["score"]
                            dl_score = dl_run_ranking[doc]
                            new_score = weight * ir_score + (1 - weight) * dl_score
                            new_ranking[doc] = new_score

                        ranking = OrderedDict(sorted(new_ranking.items(), key=lambda t: t[1], reverse=True))
                        ranking_results = evaluate_ranking(ranking, run["gnd_trh"])

                        ranking_info, ranking_results = get_info_to_ranking_results(ranking, ranking_results,
                                                                                    run, dl_model, ir_model,
                                                                                    best_weights_name, setting)

                        all_new_results.append(ranking_results)
                        all_new_rankings.append(ranking_info)

    print("--- %s seconds ---" % (time.time() - start_time))

    print("Writing data")

    pd.read_json(json.dumps(all_new_results)).to_csv(os.path.join(results_out_path, 'all_results.csv'),
                                                     index=False, sep=";")
    write_json_line_by_line(all_new_rankings, os.path.join(rankings_out_path, 'all_rankings.json'))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_prep.ipynb (unless otherwise specified).

__all__ = ['get_rand_imgs', 'vid_from_frames', 'Video', 'VideoDataset', 'get_rico_imgs', 'read_video_data',
           'get_non_duplicate_corpus', 'generate_setting2', 'get_all_texts']

# Cell
import concurrent.futures
import csv
import cv2
import ffmpeg
import json
import ntpath
import numpy
import os
import pprint
import pytesseract
import random

import matplotlib.pyplot as plt
import more_itertools as mit
import pandas as pd

from collections import defaultdict, OrderedDict
from pathlib import Path
from PIL import Image
from .utils import *
from shutil import copyfile
from tqdm.auto import tqdm

# Cell
def get_rand_imgs(vid_path, max_msecs, n = 10):
    vid = cv2.VideoCapture(str(vid_path))

    imgs = []
    while len(imgs) < n:
        msec = random.randrange(1_000, max_msecs, 1_000)
        vid.set(cv2.CAP_PROP_POS_MSEC, msec)

        success, img = vid.read()
        if success:
            imgs.append(img)

    return imgs

# Cell
def vid_from_frames(frames, output = None, fr = 30):
    """Generate video from list of frame paths."""
    if not output: output = frames.parent

    try:
        stream = ffmpeg.input(frames/'%04d.jpg')
        stream = ffmpeg.output(stream, str(output/"gen_vid.mp4"), r = fr)
        out, err = ffmpeg.run(stream)
    except Exception as e:
        print("Error occured:", e)

# Cell
class Video:
    def __init__(self, vid_path, fr = None, overwrite = False):
        self.video = cv2.VideoCapture(str(vid_path))
        self.vid_path = vid_path
        self.fr = eval(ffmpeg.probe(vid_path)["streams"][0]["avg_frame_rate"])
        if fr is not None:
            self.fr = fr
            self.vid_path = self._fix_framerate(vid_path, fr, overwrite)

    def show_frame(self, i):
        plt.imshow(self[i])
        plt.show()

    def _fix_framerate(self, vid_path, fr, overwrite):
        """
            Fixes each video in the list of video paths to a certain frame rate.
        """
        output_path = str(vid_path) if overwrite else str(vid_path.parent/f'{vid_path.stem}_fixed_{fr}.mp4')
        stream = ffmpeg.input(vid_path)
        stream = ffmpeg.output(stream, output_path, r = fr)
        stream = ffmpeg.overwrite_output(stream)
        out, err = ffmpeg.run(stream)

        return Path(output_path)

    def __len__(self):
        return int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))

    def __getitem__(self, i):
        if i >= len(self) or i < 0:
            raise Exception(f'Frame index is not in the proper range (0, {len(self) - 1}).')
        self.video.set(cv2.CAP_PROP_POS_FRAMES, i)
        suc, frame = self.video.read()
        if not suc: return None
        return Image.fromarray(frame)

# Cell
class VideoDataset:
    def __init__(self, videos):
        self.videos = videos
        self.labels = None
        self.data = None

    def label_from_paths(self):
        self.labels = defaultdict(
            lambda: defaultdict(dict)
        )
        for vid in self.videos:
            self.labels[vid.vid_path.parent.parent.name][vid.vid_path.parent.name][vid.vid_path.parent.parent.parent.name] = vid

        return self

    def get_labels(self):
        return list(self.labels.keys())

    @staticmethod
    def from_path(path, extract_frames = False, fr = None, overwrite = False):
        videos = []
        fixed_vid_paths = sorted(path.rglob(f"*fixed_{fr}.mp4"))
        if len(fixed_vid_paths) > 0:
            for vid_path in fixed_vid_paths:
                videos.append(Video(vid_path, overwrite = overwrite))
        else:
            vid_paths = sorted(path.rglob('*.mp4'))
            for vid_path in vid_paths:
                videos.append(Video(vid_path, fr = fr, overwrite = overwrite))

        return VideoDataset(videos)

    def __getitem__(self, label):
        return self.labels[label]

# Cell
def get_rico_imgs(path, n = None):
    rico_path = path/'rico-images/data'
    img_paths = sorted(rico_path.glob('*.jpg'))
    if n == None: n = len(img_paths)

    return [Image.open(img) for img in random.sample(img_paths, n)]

# Cell
def read_video_data(file_path):
    with open(file_path) as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter=';')
        data = list(csv_reader)

        idx_data = {}
        apps, bugs, app_bugs, users = set(), set(), set(), set()
        for row in data:
            app = row['app']
            bug = row['bug']
            app_bug = row['app_bug']
            user = row['final_assignment']

            apps.add(app)
            bugs.add(bug)
            app_bugs.add(app_bug)
            users.add(user)

            if app not in idx_data:
                idx_data[app] = {}
            bugs_for_app = idx_data[app]
            if bug not in bugs_for_app:
                bugs_for_app[bug] = []
            users_for_bug = bugs_for_app[bug]
            users_for_bug.append((user, app_bug + "-" + user))

    result = {
        'idx_data': idx_data,
        'apps': apps,
        'bugs': bugs,
        'app_bugs': app_bugs,
        'users': users,
        'data': data,
    }
    return result

def get_non_duplicate_corpus(bugs, bug_idx, app_bugs, br_idx, bugs_to_exclude=[]):
    if bug_idx == 0:
        other_bugs = bugs[bug_idx + 1:len(bugs)]
    elif bug_idx == len(bugs) - 1:
        other_bugs = bugs[0:bug_idx]
    else:
        other_bugs = bugs[bug_idx + 1:len(bugs)]
        other_bugs.extend(bugs[0:bug_idx])

    assert len(other_bugs) == 9, "The list of non-duplicate bugs is different than 9"
    # print(other_bugs)
    bug_reports = []
    for bug in other_bugs:
        if bug not in bugs_to_exclude:
            bug_report = app_bugs[bug][br_idx]
            bug_reports.append(bug_report[1])
    return bug_reports

def generate_setting2(data, out_path):
    Path(out_path).mkdir(parents=True, exist_ok=True)
    apps = data['apps']
    idx_data = data['idx_data']

    retrieval_runs = []
    run_id = 1

    # for each app
    for app in apps:

        app_data = idx_data[app]
        pprint.pprint(app_data)

        # for each bug
        bugs = list(idx_data[app].keys())
        for bug_idx in range(len(bugs)):
            bug = bugs[bug_idx]
            bug_reports = idx_data[app][bug]

            # for each other bug
            other_bugs_idxes = [i for i in range(len(bugs)) if i != bug_idx]
            for bug_idx2 in other_bugs_idxes:

                next_bug = bugs[bug_idx2]
                next_bug_reports = idx_data[app][next_bug]

                # for each bug report
                for br_idx in range(3):
                    query = bug_reports[br_idx][1]

                    duplicate_corpus = [bug_reports[(br_idx + 1) % 3][1], bug_reports[(br_idx + 2) % 3][1]]
                    ground_truth = duplicate_corpus.copy()
                    duplicate_corpus.extend([l[1] for l in next_bug_reports])

                    # for each user
                    for br_idx2 in range(3):
                        # get the non-duplicate corpus for each user and all other bugs except current one
                        non_duplicate_corpus = get_non_duplicate_corpus(bugs, bug_idx, app_data, br_idx2, [next_bug])

                        retrieval_job = {
                            'run_id': run_id,
                            'query': query,
                            'corpus_size': len(duplicate_corpus) + len(non_duplicate_corpus),
                            'dup_corpus': duplicate_corpus,
                            'non_dup_corpus': non_duplicate_corpus,
                            'gnd_trh': ground_truth
                        }
                        run_id += 1
                        retrieval_runs.append(retrieval_job)

    write_json_line_by_line(retrieval_runs, out_path/'setting2.json')

# Cell
def get_all_texts(vid_ds, out_path, fps):
    Path(out_path).mkdir(parents=True, exist_ok=True)

    video_output_path = os.path.join(out_path, "text_" + str(fps))
    Path(video_output_path).mkdir(parents=True, exist_ok=True)

    videos = [vid.vid_path for vid in vid_ds.videos]
    for video_path in videos:
        video_path_obj = Path(video_path)

        file_name = ntpath.basename(video_path).split(".")[0]
        video_name = file_name + "-" + str(video_path_obj.parent.parent.parent.stem)

        frame_path = os.path.join(out_path, "frames_" + str(fps), video_name)
        Path(frame_path).mkdir(parents=True, exist_ok=True)

        frames = find_file("*.jpeg", frame_path)
        if not frames:
            extract_frames(video_path_obj, frame_path, fps)
        frames = find_file("*.jpeg", frame_path)

        frames_text = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=12) as executor:
            futures = []
            for frame in frames:
                futures.append(executor.submit(process_frame, frame))
            for future in concurrent.futures.as_completed(futures):
                frames_text.append(future.result())

        frames_text = sorted(frames_text, key=lambda t: t["f"])

        video_name = video_name.replace("_fixed_30", "")
        out_file = os.path.join(video_output_path, video_name + '.json')
        write_json_line_by_line(frames_text, out_file)

        print("done: " + video_name)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_utils.ipynb (unless otherwise specified).

__all__ = ['write_json_line_by_line', 'read_csv_to_dic_list', 'read_json', 'read_json_line_by_line', 'find_file',
           'write_csv_from_json_list', 'group_dict', 'load_settings', 'get_grayscale', 'remove_noise', 'thresholding',
           'thresholding_med', 'dilate', 'erode', 'opening', 'canny', 'deskew', 'match_template', 'extract_text',
           'preprocess_img', 'extract_frames', 'process_frame']

# Cell
import csv
import cv2
import fnmatch
import json
import ntpath
import os
import pytesseract

import numpy as np
import pandas as pd

from itertools import groupby

# Cell
def write_json_line_by_line(data, file_path):
    with open(file_path, 'w') as dest_file:
        for record in data:
            print(json.dumps(record), file=dest_file)


def read_csv_to_dic_list(file_path):
    data = []
    with open(file_path) as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter=';')
        for item in csv_reader:
            data.append(item)
    return data


def read_json(file_path):
    with open(file_path) as file:
        return json.load(file)


def read_json_line_by_line(file_path):
    data = []
    with open(file_path) as sett_file:
        for item in map(json.loads, sett_file):
            data.append(item)
    return data


def find_file(pattern, path):
    result = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                if 'fix' not in name:
                    result.append(os.path.join(root, name))
    return result


def write_csv_from_json_list(data, output_path):
    pd.read_json(json.dumps(data)).to_csv(output_path, index=False, sep=";")


def group_dict(data, lambda_expr):
    result = {}
    data.sort(key=lambda_expr)
    for k, v in groupby(data, key=lambda_expr):
        result[k] = list(v)
    return result


def load_settings(path):
    settings_files = find_file("*.json", path)

    all_settings = {}
    for file_path in settings_files:
        setting_name = ntpath.basename(file_path).split(".")[0]
        all_settings[setting_name] = []
        with open(file_path) as sett_file:
            for retrieval_run in map(json.loads, sett_file):
                all_settings[setting_name].append(retrieval_run)

    return all_settings

# Cell

# get grayscale image
def get_grayscale(image):
    return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)


# noise removal
def remove_noise(image):
    return cv2.medianBlur(image, 5)


# thresholding
def thresholding(image):
    return cv2.threshold(image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

def thresholding_med(image):
    return cv2.threshold(cv2.medianBlur(image, 3), 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

# dilation
def dilate(image):
    kernel = np.ones((5, 5), np.uint8)
    return cv2.dilate(image, kernel, iterations=1)


# erosion
def erode(image):
    kernel = np.ones((5, 5), np.uint8)
    return cv2.erode(image, kernel, iterations=1)


# opening - erosion followed by dilation
def opening(image):
    kernel = np.ones((1, 1), np.uint8)
    return cv2.morphologyEx(image, cv2.MORPH_OPEN, kernel)


# canny edge detection
def canny(image):
    return cv2.Canny(image, 100, 200)


# skew correction
def deskew(image):
    coords = np.column_stack(np.where(image > 0))
    angle = cv2.minAreaRect(coords)[-1]
    if angle < -45:
        angle = -(90 + angle)
    else:
        angle = -angle
    (h, w) = image.shape[:2]
    center = (w // 2, h // 2)
    M = cv2.getRotationMatrix2D(center, angle, 1.0)
    rotated = cv2.warpAffine(image, M, (w, h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
    return rotated


# template matching
def match_template(image, template):
    return cv2.matchTemplate(image, template, cv2.TM_CCOEFF_NORMED)

# Cell
def extract_text(img):
    custom_config = r'--oem 3 --psm 3'
    return pytesseract.image_to_string(img, config=custom_config)


def preprocess_img(image):
    prep_img = get_grayscale(image)
    prep_img = opening(prep_img)
    prep_img = thresholding_med(prep_img)
    return prep_img


def extract_frames(video_path, out_path, fps):
    ffmpeg_command = f'ffmpeg -i {video_path} -vf "fps={fps}" {out_path}/%04d.jpeg'
    os.system(ffmpeg_command)


def process_frame(frame):
    image_frame = cv2.imread(frame)
    prep_image = preprocess_img(image_frame)
    text = extract_text(prep_image)

    frame_name = ntpath.basename(frame).split(".")[0]
    record = {"f": frame_name, "txt": text}
    return record


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_rand_imgs": "00_prep.ipynb",
         "vid_from_frames": "00_prep.ipynb",
         "Video": "00_prep.ipynb",
         "VideoDataset": "00_prep.ipynb",
         "get_rico_imgs": "00_prep.ipynb",
         "read_video_data": "00_prep.ipynb",
         "get_non_duplicate_corpus": "00_prep.ipynb",
         "generate_setting2": "00_prep.ipynb",
         "get_all_texts": "00_prep.ipynb",
         "Extractor": "01_features.ipynb",
         "SIFTExtractor": "01_features.ipynb",
         "CNNExtractor": "01_features.ipynb",
         "imagenet_normalize_transform": "03_model.ipynb",
         "get_transforms": "01_features.ipynb",
         "SimCLRExtractor": "01_features.ipynb",
         "gen_vcodebook": "01_features.ipynb",
         "gen_codebooks": "01_features.ipynb",
         "get_df": "01_features.ipynb",
         "get_bovw": "01_features.ipynb",
         "extract_features": "01_features.ipynb",
         "new_get_bovw": "01_features.ipynb",
         "calc_tf_idf": "02_eval.ipynb",
         "cosine_similarity": "02_eval.ipynb",
         "hit_rate_at_k": "02_eval.ipynb",
         "mean_reciprocal_rank": "02_eval.ipynb",
         "r_precision": "02_eval.ipynb",
         "precision_at_k": "02_eval.ipynb",
         "average_precision": "02_eval.ipynb",
         "mean_average_precision": "02_eval.ipynb",
         "recall_at_k": "02_eval.ipynb",
         "rank_stats": "02_eval.ipynb",
         "evaluate": "02_eval.ipynb",
         "get_eval_results": "02_eval.ipynb",
         "evaluate_ranking": "02_eval.ipynb",
         "sift_frame_sim": "03_model.ipynb",
         "simclr_frame_sim": "03_model.ipynb",
         "SimCLRDataset": "03_model.ipynb",
         "get_train_transforms": "03_model.ipynb",
         "get_val_transforms": "03_model.ipynb",
         "NTXEntCriterion": "03_model.ipynb",
         "SimCLRModel": "03_model.ipynb",
         "flatten_dict": "04_approach.ipynb",
         "gen_extracted_features": "04_approach.ipynb",
         "gen_tfidfs": "04_approach.ipynb",
         "gen_bovw_similarity": "04_approach.ipynb",
         "fuzzy_LCS": "04_approach.ipynb",
         "gen_lcs_similarity": "04_approach.ipynb",
         "fix_sims": "04_approach.ipynb",
         "sort_rankings": "04_approach.ipynb",
         "approach": "04_approach.ipynb",
         "compute_sims": "04_approach.ipynb",
         "logger": "05_cli.ipynb",
         "URLs": "05_cli.ipynb",
         "download": "05_cli.ipynb",
         "VWORDS": "05_cli.ipynb",
         "N_IMGS": "05_cli.ipynb",
         "N_FRAMES_TO_KEEP": "05_cli.ipynb",
         "FPS": "05_cli.ipynb",
         "BEST_DL_MODELS": "05_cli.ipynb",
         "BEST_IR_MODELS": "05_cli.ipynb",
         "BEST_MODEL_CONFIGS": "05_cli.ipynb",
         "reproduce": "05_cli.ipynb",
         "tango": "05_cli.ipynb",
         "write_json_line_by_line": "07_utils.ipynb",
         "read_csv_to_dic_list": "07_utils.ipynb",
         "read_json": "07_utils.ipynb",
         "read_json_line_by_line": "07_utils.ipynb",
         "find_file": "07_utils.ipynb",
         "write_csv_from_json_list": "07_utils.ipynb",
         "group_dict": "07_utils.ipynb",
         "load_settings": "07_utils.ipynb",
         "get_grayscale": "07_utils.ipynb",
         "remove_noise": "07_utils.ipynb",
         "thresholding": "07_utils.ipynb",
         "thresholding_med": "07_utils.ipynb",
         "dilate": "07_utils.ipynb",
         "erode": "07_utils.ipynb",
         "opening": "07_utils.ipynb",
         "canny": "07_utils.ipynb",
         "deskew": "07_utils.ipynb",
         "match_template": "07_utils.ipynb",
         "extract_text": "07_utils.ipynb",
         "preprocess_img": "07_utils.ipynb",
         "extract_frames": "07_utils.ipynb",
         "process_frame": "07_utils.ipynb",
         "execute_retrieval_run": "08_combo.ipynb",
         "run_settings": "08_combo.ipynb",
         "write_results": "08_combo.ipynb",
         "write_rankings": "08_combo.ipynb",
         "convert_results_format": "08_combo.ipynb",
         "get_info_to_ranking_results": "08_combo.ipynb",
         "tango_combined": "08_combo.ipynb"}

modules = ["prep.py",
           "features.py",
           "eval.py",
           "model.py",
           "approach.py",
           "cli.py",
           "utils.py",
           "combo.py"]

doc_url = "https://ncoop57.github.io/tango/"

git_url = "https://github.com/ncoop57/tango/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_features.ipynb (unless otherwise specified).

__all__ = ['Extractor', 'SIFTExtractor', 'CNNExtractor', 'imagenet_normalize_transform', 'get_transforms',
           'SimCLRExtractor', 'gen_vcodebook', 'gen_codebooks', 'get_df', 'get_bovw', 'extract_features',
           'new_get_bovw', 'calc_tf_idf']

# Cell
import cv2
import pickle
# import ffmpeg
import random
import time
import torch

import numpy as np

from abc import ABC, abstractmethod
from pathlib import Path
from PIL import Image
from .prep import *
from tqdm.auto import tqdm
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import cosine_similarity

# Cell
class Extractor(ABC):

    def __init__(self, extractor):
        self.extractor = extractor
        super().__init__()

    @abstractmethod
    def extract(self, img):
        pass

# Cell
class SIFTExtractor(Extractor):

    '''Exposed SIFTExtractor class used for retrieving features.'''

    def extract(self, img):
        '''Given an image, extract features using SIFT. Returns the feature vector.'''
        img = np.array(img)
        _, features = self.extractor.detectAndCompute(img, None)
        return features

class CNNExtractor(Extractor):

    '''Exposed CNNExtractor class used for retrieving features.'''

    def extract(self, img):
        '''Given an image, extract features from the layers of a CNN. Returns the feature vector.'''

        return self.extractor.getFeatures(img)

# Cell
from torchvision import transforms

def imagenet_normalize_transform():
    return transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

def get_transforms(size=224):
    return transforms.Compose([
        transforms.Resize(size=(size, size)),
        transforms.ToTensor(),
        imagenet_normalize_transform()
    ])

class SimCLRExtractor(Extractor):

    '''Exposed CNNExtractor class used for retrieving features.'''
    def __init__(self, extractor):
        super().__init__(extractor)
        self.transforms = get_transforms()

    def extract(self, img):
        '''Given an image, extract features from the layers of a CNN. Returns the feature vector.'''
        img = self.transforms(img).float()
        img = img.unsqueeze(0)
        return self.extractor(img).detach().numpy()

# Cell
def gen_vcodebook(path, img_paths, model_name, extractor, vwords):
    """
        Constructs a visual codebook based on the given images.
        You can change vwords to increase the vocabulary of the codebook.
    """
    fname = path/f'models/features/{model_name}/{len(img_paths)}n_features.pkl'
    features_list = []
    feature_time = 0
    if not fname.is_file():
        feature_start = time.time()
        for img in tqdm(img_paths):
            features = extractor.extract(Image.open(img))
            if features is None: continue
            features_list.extend(features)

        features_list = np.asarray(features_list)
        pickle.dump(features_list, open(fname, 'wb'), protocol=4)
        feature_end = time.time()
        feature_time = feature_end - feature_start
        fname = path/f'models/features/{model_name}/{len(img_paths)}n_features_elapsed_time.txt'
        with open(fname, 'w') as f:
            f.write(f'{feature_time}')
    else:
        features_list = pickle.load(open(fname, 'rb'))
        with open(fname, 'r') as f:
            feature_time = [float(x) for x in f][0]
    codebooks = []
    for vw in vwords:
        cb_start = time.time()
        codebook = KMeans(n_clusters = vw)
        codebook.fit(features_list)
        cb_end = time.time()
        codebooks.append((cb_end - cb_start + feature_time, codebook))

    return codebooks

# Cell
def gen_codebooks(path, models, vwords, samples = 15_000):
    rico_path = path/'rico-images/data'
    img_paths = random.sample(list(rico_path.glob('*.jpg')), samples)
    for model_name, model in models:
        codebooks = gen_vcodebook(path, img_paths, model_name, model, vwords)
        for (cb_time, codebook), vw in zip(codebooks, vwords):
            fname = path/f'models/codebooks/{model_name}/cookbook_{model_name}_{vw}vw.model'
            pickle.dump(codebook, open(fname, 'wb'))
            fname = path/f'models/codebooks/{model_name}/cookbook_{model_name}_{vw}vw_elapsed_time.txt'
            with open(fname, 'w') as f:
                f.write(f'{cb_time}')

# Cell
def get_df(imgs, extractor, codebook, vwords):
    """Generates the document frequency for the visual words"""
    arr = []
    for img in imgs:
        features = extractor.extract(img)
        vw = codebook.predict(features)
        arr.extend(vw)
    arr = np.asarray(arr)

    return np.histogram(arr, bins = range(vwords + 1))

# Cell
def get_bovw(vid_path, extractor, codebook, vwords, n = None):
    """Generates the bag of visual words (bovw) for an entire video."""
    vid = cv2.VideoCapture(str(vid_path))
    if n is None: n = vid.get(cv2.CAP_PROP_FRAME_COUNT)

    # checks whether frames were extracted
    success = 1
    bovw = np.array([])
    for i in progress_bar(range(n)):
        # vid object calls read
        # function extract frames
        success, img = vid.read()
        if success:
            features = extractor.extract(img)
            vw = codebook.predict(features)
            bovw = np.concatenate((bovw, vw))

    hist = np.histogram(bovw, bins = range(vwords + 1))[0]
    return hist, bovw

# Cell
def extract_features(vid, extractor, fps = 30, frames_to_keep = 1):
    extracted_features = []
    for i in range(0, len(vid), int(fps / frames_to_keep)):
        img = vid[i]
        if not img: continue
        extracted_features.append(extractor.extract(img))

    return extracted_features

# Cell
def new_get_bovw(features, codebook, vwords):
    bovw = []
    for f in features:
        vw = codebook.predict(f)
        bovw.extend(vw)

    bovw = np.array(bovw)
    bovw = np.histogram(bovw, bins = range(vwords + 1))[0]
    return bovw

# Cell
def calc_tf_idf(tfs, dfs):
    tf_idf = np.array([])
    for tf, df in zip(tfs, dfs):
        tf = tf / np.sum(tfs)
        idf = np.log(len(tfs) / (df + 1))
        tf_idf = np.append(tf_idf, tf * idf)

    return tf_idf